/*--- file keyman.c -- Manage public keys
 *
 *  Mark Riordan  20 May 1992
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "global.h"
#include "rsaref.h"
#include "md5.h"
#include "ripem.h"
#include "keyfield.h"
#ifdef USE_DDES
#include "ddes.h"
#else
#include "des.h"
#endif
#include "keymanpr.h"
#include "strutilp.h"
#include "derkeypr.h"
#include "hexbinpr.h"
#include "ripemsop.h"
#include "pubinfop.h"
#include "keyderpr.h"
#include "bfstream.h"
#include "rdwrmsgp.h"
#include "certder.h"
#include "certutil.h"

#include "bemparse.h"
#include "p.h"

#define DER_SEQ 0x30
#define RECSIZE 3000
#define MAXKEYBYTES 1024

#define MAX_PRENCODE_BYTES 48

static char *GetKeyBytesFromFile
  P((char *, TypFile *, char *, BOOL *, unsigned char **, unsigned int *,
     RIPEMInfo *));
static char *GetUserRecordFromFile
  P((char *, TypFile *, unsigned int, char *, BOOL *, RIPEMInfo *));

/*  Get the public key of a user from a PublicKeyInfo field (as opposed
 *    to a certificate.  This is provided only for compatibility with
 *    RIPEM 1.1.
 *
 *  Entry:  user      is the user's email address.
 *          source      tells us where to look.
 *          ripemInfo is used to access debug and debugStream.
 * 
 *   Exit:  key     is the key (if found).
 *          found     is TRUE if we found the key.
 *          Returns an error message if something goes wrong more 
 *            serious than not being able to find the key.
 */
char *GetUnvalidatedPublicKey (user, source, key, found, ripemInfo)
char *user;
TypKeySource *source;
R_RSA_PUBLIC_KEY *key;
BOOL *found;
RIPEMInfo *ripemInfo;
{
  unsigned char *bytes = NULL;
  char *errorMessage=NULL;
  int sour;
  unsigned int derLen = 0;
  static BOOL server_ok=TRUE;
  char rec[RECSIZE];
  char coded_bytes[MAXKEYBYTES];
  unsigned char key_bytes[MAXKEYBYTES];
#define MAXDIGESTSIZE 36
  char computed_hex_digest[MAXDIGESTSIZE], records_hex_digest[MAXDIGESTSIZE];
  TypFile *fptr;

  *found = FALSE;
  
  for(sour=0; !*found && sour<MAX_KEY_SOURCES; sour++)
    switch(source->origin[sour]) {
    /* Search the key file */
  case KEY_FROM_FILE:
    FORLIST(&(source->filelist));
      fptr = (TypFile *)dptr;
      if(!fptr->stream) {
        sprintf(ripemInfo->errMsgTxt,"Can't open public key file \"%s\".",
                fptr->filename);
        errorMessage = ripemInfo->errMsgTxt;
        continue;
      }

      /* Get the user record for this user from the file. */
      errorMessage = GetUserRecordFromFile
        (user,fptr,RECSIZE,rec,found,ripemInfo);
      if(*found) goto decodekey;
    ENDFORLIST;
    break;

    /* Get key from server. */
  case KEY_FROM_SERVER:
    if(server_ok) {
      errorMessage = GetUserRecordFromServer
        (user,source, rec,RECSIZE,&server_ok,found,ripemInfo);
      goto decodekey;
    }
            
    /* Get key from finger. */
  case KEY_FROM_FINGER:
    errorMessage = GetUserRecordFromFinger (user,rec,RECSIZE,found,ripemInfo);
  decodekey:
    if(ripemInfo->debug>1) {
      if(errorMessage) {
        fprintf(ripemInfo->debugStream,
                "Error retrieving key for %s from server: %s\n",
                user,errorMessage);
      } else if(*found) {
        fprintf(ripemInfo->debugStream,
                "Found %s's public key record:\n%s\n",user,rec);
      } else {
        fprintf(ripemInfo->debugStream,
                "Couldn't find %s's public key record.\n", user);
      }
    }
    if(*found) {
      CrackKeyField(rec,PUBLIC_KEY_FIELD,coded_bytes,MAXKEYBYTES);
      if(ripemInfo->debug>1) {
        fprintf(ripemInfo->debugStream,"Coded pub key=\"%s\"\n",coded_bytes);
      }
      /* CrackKeyField skips all whitespace and null-terminates the
         coded_bytes.  key_bytes is the same size as coded_bytes so
         the smaller decoded bytes will fit. */
      R_DecodePEMBlock (key_bytes, &derLen, coded_bytes, strlen (coded_bytes));
      bytes = key_bytes;
    }
            
    break;

  default:
    break;   /* XXX */
  }
  if(*found) {
    /* Convert key bytes to public key structure format. */
    if(DERToPubKey(bytes,key)) {
      /* Conversion didn't work. */
      sprintf(ripemInfo->errMsgTxt,"Error parsing public key for %s.",user);
      errorMessage = ripemInfo->errMsgTxt;
    } else {
      /* Conversion from BER format worked OK.
       * Now check this public key against the enclosed digest.
       */
      MakeHexDigest((unsigned char *)bytes,derLen,computed_hex_digest);
      if(CrackKeyField(rec,PUBLIC_KEY_DIGEST_FIELD,
                       records_hex_digest,MAXDIGESTSIZE)) {
        if(ripemInfo->debug > 2) {
          fprintf(ripemInfo->debugStream,
                  "der len=%u\nComputed  MD5 of %s's pubkey=%s\n",derLen,
                  user, computed_hex_digest);
          fprintf(ripemInfo->debugStream,
                  "Retrieved MD5 of %s's pubkey=%s\n",user,
                  records_hex_digest);
        }
        if(strcmp(computed_hex_digest,records_hex_digest)) {
          sprintf(ripemInfo->errMsgTxt,
                  "Public key of '%s' is garbled--digest does not match.",
                  user);
          errorMessage = ripemInfo->errMsgTxt;
        }
      } else {
        fprintf(stderr,"Warning--could not find %s's key digest.\n",
                user);
      }
    }
  } 
  return errorMessage;
}

/*--- function GetPrivateKey -------------------------------------
 *
 *  Get the private key of a user.
 *  RIPEMInfo is used for debug and debugStream.
 */
char *
GetPrivateKey (user, source, key, password, passwordLen, ripemInfo)
char *user;
TypKeySource *source;
R_RSA_PRIVATE_KEY *key;
unsigned char *password;
unsigned int passwordLen;
RIPEMInfo *ripemInfo;
{
  unsigned char *bytes;
  unsigned int nbytes;
  char *errorMessage=NULL;
  unsigned char salt[SALT_SIZE];
  unsigned char *enc_key;
  BOOL found;
  unsigned int enc_key_len, iter_count;
  int digest_alg;
  unsigned int num_der_bytes;
  int sour;
  TypFile *fptr;

  for(sour=0; sour<MAX_KEY_SOURCES; sour++)
    switch(source->origin[sour]) {
    /* Search the key file */
  case KEY_FROM_FILE:
    FORLIST(&(source->filelist));
      fptr = (TypFile *)dptr;
      if(!fptr->stream) {
        sprintf(ripemInfo->errMsgTxt,"Can't open private key file \"%s\".",
                fptr->filename);
        errorMessage = ripemInfo->errMsgTxt;
        continue;
      }
      /* Rewind in order to search from the beginning. */
      fseek (fptr->stream, 0L, 0);
      errorMessage = GetKeyBytesFromFile
        (user,fptr, PRIVATE_KEY_FIELD,&found,&bytes,&nbytes,ripemInfo);
      if(!errorMessage & !found) {
        errorMessage = "Can't find user's private key.";
      }
      if(found) {
        errorMessage = NULL;
        break;
      }
    ENDFORLIST;
    break;

  default:
    break;   /* XXX */
  }

  /* We now have the DER-encoded encrypted private key structure.
   * First, decode it to obtain the encryption algorithm parameters
   * and the actual bytes of the encrypted key.
   */
  if(!errorMessage) {
    if(ripemInfo->debug>1) {
      fprintf(ripemInfo->debugStream,"Obtained %u byte encrypted private key for %s:\n",
              nbytes,user);
      BEMParse(bytes,ripemInfo->debugStream);
    }
    enc_key = (unsigned char *) malloc(nbytes);
    if(!enc_key) return ERR_MALLOC;
    if(DERToEncryptedPrivKey (bytes,nbytes, &digest_alg,salt,&iter_count,
                              enc_key,&enc_key_len)) {
      return "Error decoding encrypted private key.";
    }
    
    /* Now decrypt the encrypted key.
     */
    if(pbeWithMDAndDESWithCBC
       (FALSE, digest_alg, enc_key, enc_key_len, password, passwordLen,
        salt, iter_count, &num_der_bytes) != 0) {
      return "Can't decrypt private key.";
    }
    
    /* We have the plaintext private key in DER format.
     * Check to make sure it looks as if it was decrypted OK.
     * Then Decode to RSAREF format.
     */
    if(enc_key[0] != DER_SEQ) {
      return "Private key could not be decrypted with this password.";
    }

    if(DERToPrivKey(enc_key,key)) {
      sprintf(ripemInfo->errMsgTxt,"Error parsing private key for %s.",user);
      errorMessage = ripemInfo->errMsgTxt;
    }
    
    if(ripemInfo->debug>1) {
      fprintf(ripemInfo->debugStream,"Dump of decrypted private key:\n");
      DumpPrivKey(key, ripemInfo->debugStream);
    }

  }
  return errorMessage;
}

/*--- function GetKeyBytesFromFile ---------------------------------
 *  
 *  Obtain the value of a field from a flat file.
 *  This searches for the next record with the given keyFieldName.
 *  The calling routine must rewind the file to begin a new search.
 *  The file is ASCII newline-delimited and consists of
 *  keys of form:   Key: value
 *  and fields of form:
 *     Fieldname:
 *        value...  (RFC1113 encoded)
 *
 *  Entry:  user        is the key value.  
 *          fileptr     contains the stream pointing at the file.
 *          keyFieldName  is the name of the field, e.g. PublicKeyInfo.
 *          ripemInfo is used for debug and debugStream.
 *          
 *  Exit:   keyBytes    points to the bytes as a newly allocated buffer.
 *          numBytes    is the number of bytes retrieved.
 *          found       is TRUE if we found the key.
 *          Returns an error message if an error was found (more serious
 *          than the key value not being found).
 */
static char *
GetKeyBytesFromFile
  (user,fileptr,keyFieldName,found,keyBytes,numBytes,ripemInfo)
char *user;
TypFile *fileptr;
char *keyFieldName;
BOOL *found;
unsigned char **keyBytes;
unsigned int *numBytes;
RIPEMInfo *ripemInfo;
{
#define KALLOC_INC 1080
#define VALUELEN 120
  char value[VALUELEN];
  BOOL reading=TRUE;
  unsigned char *bytes, *base;
  char *errorMessage = NULL;
  int bytesleft, alloc_size;
  unsigned int bytesInLine;
  
  *found = FALSE;
  bytes = (unsigned char *)malloc(KALLOC_INC);
  *keyBytes = bytes;
  if(!*keyBytes) return (ERR_MALLOC);
  alloc_size = bytesleft = KALLOC_INC;
  *numBytes = 0;
  
  /* Position to just after the line containing the user's name. */
  while(reading && GetFileLine(fileptr->stream,"User:",
                               value,VALUELEN)) {
    if(match(value,user)) {
      reading = FALSE;
      *found = TRUE;
    }
  }
  /* We are now in the section of the file that corresponds
   * to this user.  Position to the desired field.
   * (There aren't many fields; we do this in case there
   * are multiple "User:" fields & to account for future changes.)
   */
  if(*found) {
    *found = FALSE;
    if(PosFileLine(fileptr->stream,keyFieldName)) {
      *found = TRUE;
      /* Now read the RFC1113-encoded lines and translate
       * them to binary.
       */

      for(reading=TRUE; reading; ) {
        if(fgets(value,VALUELEN,fileptr->stream)) {
          if(WhiteSpace(value[0]) && !LineIsWhiteSpace(value)) {
            /* If we need more room in keyBytes, allocate more.
             * Reassign pointers as necessary.
             */
            if(bytesleft < MAX_PRENCODE_BYTES) {
              alloc_size += KALLOC_INC;
              base = (unsigned char *)realloc(*keyBytes,alloc_size);
              if(base) {
                *keyBytes = base;
                bytes = *keyBytes + *numBytes;
              } else {
                free(*keyBytes);
                return(ERR_MALLOC);
              }
            }
            /* Decode the line into the keyBytes buffer.
               Assume the input line has one prefix space and that the
                 decode won't return more than MAX_PRENCODE_BYTES.
               Assume each line except the last has a multiple of 4 chars.
               Also assume there is one '\n' and no spaces at the end of line.
             */
            R_DecodePEMBlock
              (bytes, &bytesInLine, (unsigned char *)(value + 1),
               strlen (value + 1) - 1);
            bytes += bytesInLine;
            *numBytes += bytesInLine;
            bytesleft -= bytesInLine;
          } else {
            reading = FALSE; /* Have reached end of field */
          }
        } else {
          if (!feof (fileptr->stream))
            errorMessage = "Error reading bytes from flat file";
          reading = FALSE;
        }
      }
    }
  }

  if(ripemInfo->debug>1) {
    if(*found) {
      fprintf(ripemInfo->debugStream,"Obtained %s for %s from file %s.\n",
              keyFieldName,user,fileptr->filename);
    } else {
      fprintf(ripemInfo->debugStream,
              "Could not find %s of user %s in file %s.\n",
              keyFieldName,user,fileptr->filename);
    } 
  }
  return errorMessage;
}

/*--- function GetUserRecordFromFile --------------------------------------
 *
 *  Get the user record for a user from a file.  A user record
 *  is a series of ASCII lines like:
 *    User: joe@bigu.edu
 *    PublicKeyInfo:
 *     MIGcMAoGBFUIAQECAgQAA4GNADCBiQKBgQDGQci5pOCGqQgW6XUYyGCcZFIyyLb7
 *     18nsKtQNjHZRODHkd+5tmHzMWp2BdFfV+CQzbMeNcdC9lC/RhLb7AgMBAAE=
 *    MD5OfPublicKey: E69AB9AA2A2697FCB5B1821DC3596345
 *
 *  Entry:  user     is the user name (email address) whose record we want.
 *          fileptr  contains the stream pointing at the file.
 *          maxBytes is the size of the buffer used to return data.
 *          ripemInfo is used for debug and debugStream.
 * 
 *  Exit:   userRec  contains the user record, if found.  It is 
 *                   zero-terminated.
 *          found    is TRUE if we found the key.
 *          Returns an error message if an error was found (more serious
 *          than the key value not being found), else NULL.
 */
static char *
GetUserRecordFromFile(user,fileptr,maxBytes,userRec,found,ripemInfo)
char *user;
TypFile *fileptr;
unsigned int maxBytes;
char *userRec;
BOOL *found;
RIPEMInfo *ripemInfo;
{
  BOOL got_next_rec, looking=TRUE;
   
  *found = FALSE;
  fseek(fileptr->stream,0L,0);  /* Rewind the file. */
      
  if(ripemInfo->debug>1) {
    fprintf(ripemInfo->debugStream,"Looking in '%s' for public key for %s.\n",
            fileptr->filename,user);
  }
  while(looking) {
    GetNextUserRecordFromFile(fileptr->stream,maxBytes,
                              userRec,&got_next_rec);
    if(got_next_rec) {
      if(FindUserInRecord(user,userRec)) {
        *found = TRUE;
        looking = FALSE;
      }
    } else {
      looking = FALSE;
    }
  }
  if(ripemInfo->debug>1) {
    if(*found) {
      fprintf(ripemInfo->debugStream,"Found %s's public key record in file.\n",
              user);
    } else {
      fprintf(ripemInfo->debugStream,"Didn't find %s's public key in file.\n",
              user);
    }
  }
  return (char *)NULL;
}

/*--- function GetNextUserRecordFromFile ----------------------------------
 *
 *  Get the next user record from a sequential file.
 *  A user record is just a sequence of lines limited by a 
 *  blank line or a line starting with "--".
 *
 *  Entry:  ustream     is the stream of the file.
 *          maxBytes    is the buffer size of userRec.
 *
 *  Exit:   userRec     is the user record, if found.
 *          found       is TRUE if we successfully retrieved a record.
 *          Returns an error message if there was a problem worse
 *          than EOF, else NULL.
 */
char *
GetNextUserRecordFromFile(ustream,maxBytes,userRec,found)
FILE *ustream;
unsigned int maxBytes;
char *userRec;
BOOL  *found;
{
#define LINELEN 256
  char line[LINELEN];
  char *errmsg=NULL;
  char *uptr=userRec;
  char *got_line;
  unsigned int mylen;
  
  *found = FALSE;
  /* Skip past leading blank lines */
  do {
    if(!fgets(line,LINELEN,ustream)) {
      if (!feof (ustream))
        errmsg = "Error in stream while reading user record";
      goto endnext;
    }
  } while(LineIsWhiteSpace(line) || strncmp(line,"---",2)==0);
  
  /* We hit a non-blank line.
   * Copy lines into the buffer until we hit EOF or blank line. 
   */
  *found = TRUE;
  do {
    mylen = (unsigned int)strlen(line);
    /* Copy this line into the buffer if there's room, 
     * else just return the truncated buffer.
     */
    if(maxBytes > mylen) {
      strcpy(uptr,line);
      uptr += mylen;
      maxBytes -= mylen;
    } else {
      goto endnext;
    }
    got_line = fgets(line,LINELEN,ustream);
  } while(got_line && !LineIsWhiteSpace(line) && strncmp(line,"---",2));
  
endnext:
  return errmsg;
}

/*--- function pbeWithMDAndDESWithCBC --------------------------------------
 *
 * Encrypt or decrypt a buffer with DES-CBC, using a key derived
 * from using a message disest (MDx) function on a password
 * and salt value.
 */
int
pbeWithMDAndDESWithCBC
  (encrypt, digestAlg, buf, numInBytes, password, passwordLen, salt,
   iterationCount, numOutBytes)
int encrypt;
int digestAlg;
unsigned char *buf;
unsigned int numInBytes;
unsigned char *password;
unsigned int passwordLen;
unsigned char *salt;
unsigned int iterationCount;
unsigned int *numOutBytes;
{
#define DIGEST_SIZE 16
  unsigned char *pass_and_salt, byte, parity, *bptr;
  MD5_CTX context;
  unsigned char digest[DIGEST_SIZE], des_key[DES_KEY_SIZE];
  unsigned char iv[DES_BLOCK_SIZE];
  unsigned int n_pad_bytes;
  unsigned int j, bit;
  
  if(digestAlg != DA_MD5) return 1;
  
  /* Create concatenation of password and salt */
  pass_and_salt = (unsigned char *) malloc(passwordLen+SALT_SIZE);
  if(!pass_and_salt) return 1;
  R_memcpy(pass_and_salt,password,passwordLen);
  R_memcpy(pass_and_salt+passwordLen,salt,SALT_SIZE);
  
  /* First iteration is a digest of password || salt */
  MD5Init(&context);
  MD5Update(&context,pass_and_salt, passwordLen+SALT_SIZE);
  MD5Final(digest,&context);
  
  free(pass_and_salt);
  
  /* Subsequent iterations are digests of the previous digest. */
  while(--iterationCount) {
    MD5Init(&context);
    MD5Update(&context,digest,DIGEST_SIZE);
    MD5Final(digest,&context);
  }
  
  /* Create the DES key by taking the first 8 bytes of the
   * digest and setting the low order bit to be an odd parity bit.
   */
  for(j=0; j<DES_KEY_SIZE; j++) {
    byte = digest[j];
    for(parity=0x01,bit=0; bit<7; bit++) {
      byte >>= 1;
      parity ^= (byte&0x01);
    }
    des_key[j] = (digest[j]&0xfe) | parity;
  }
  
  /* Create the initialization vector from the last 8 bytes of the digest */
  R_memcpy(iv,digest+DES_KEY_SIZE,DES_BLOCK_SIZE);
  
  /* Now we have the DES key and the init vector.
   * Do the encrypt or decrypt.
   */

  if(encrypt) {
    /* Pad the last block of the buffer with 1 to 8 bytes of
     * the value 01 or 0202 or 030303 or...
     */

    n_pad_bytes = DES_BLOCK_SIZE - numInBytes%DES_BLOCK_SIZE;
    for(bptr=buf+numInBytes,j=0; j<n_pad_bytes; j++,bptr++) {
      *bptr = n_pad_bytes;
    }
    *numOutBytes = numInBytes+n_pad_bytes;
    
    DESWithCBC(encrypt,buf,*numOutBytes,des_key,iv);
    
  } else {
    /* Do the decryption */
    if(numInBytes%DES_BLOCK_SIZE) return 1;
    DESWithCBC(encrypt,buf,numInBytes,des_key,iv);
    n_pad_bytes = buf[numInBytes-1];
    *numOutBytes = numInBytes - n_pad_bytes;
  }
  
  return 0;
}


/*--- function DESWithCBC ------------------------------------------
 *
 * Encrypt or decrypt a buffer with DES with Cipher Block Chaining.
 *
 *  Entry:  encrypt  is TRUE to encrypt, else decrypt.
 *          buf      is the beginning of the buffer.
 *          numBytes is the number of bytes to encrypt/decrypt.
 *                   It is rounded up to a multiple of 8.
 *          key      is the 8-byte key.
 *          iv       is the initialization vector.  (We pretend
 *                   it is the output from the previous round
 *                   of encryption.)
 *
 *  Exit:   buf      has been encrypted/decrypted.
 */
void
DESWithCBC(encrypt,buf,numBytes,key,iv)
int encrypt;
unsigned char *buf;
unsigned int numBytes;
unsigned char *key;
unsigned char *iv;
{
#ifdef USE_DDES
  int mode = !encrypt, count;
  unsigned char my_iv[DES_BLOCK_SIZE], save_iv[DES_BLOCK_SIZE];
  unsigned char *source, *targ;
  unsigned int block_cnt = (numBytes+DES_BLOCK_SIZE-1)/DES_BLOCK_SIZE;
  
  deskey(key,mode);
  R_memcpy(my_iv,iv,DES_BLOCK_SIZE);
  
  if(encrypt) {
    while(block_cnt--) {
      for(targ=buf,source=my_iv,count=DES_BLOCK_SIZE; count; count--) {
        *(targ++) ^= *(source++);
      }
      des(buf,buf);
      R_memcpy(my_iv,buf,DES_BLOCK_SIZE);
      buf += DES_BLOCK_SIZE;
    }
  } else {
    while(block_cnt--) {
      R_memcpy(save_iv,buf,DES_BLOCK_SIZE);
      des(buf,buf);
      for(targ=buf,source=my_iv,count=DES_BLOCK_SIZE; count; count--) {
        *(targ++) ^= *(source++);
      }
      R_memcpy(my_iv,save_iv,DES_BLOCK_SIZE);
      buf += DES_BLOCK_SIZE;
    }
  }
#else
  DES_CBC_CTX context;
  unsigned int len;
  
  len = (numBytes+DES_BLOCK_SIZE-1)/DES_BLOCK_SIZE;
  len *= DES_BLOCK_SIZE;
  
  DES_CBCInit(&context, key, iv, encrypt);
  DES_CBCUpdate(&context, buf, buf, len);
  R_memset ((POINTER)&context, 0, sizeof (context));
#endif
}

/*--- function DumpPubKey ------------------------------------------
 *
 */
void
DumpPubKey(pubKey, stream)
R_RSA_PUBLIC_KEY *pubKey;
FILE *stream;
{
  fprintf(stream,"Dump of %d bit key:\n",pubKey->bits);
  fprintf(stream,"    Mod=");
  DumpBigNum(pubKey->modulus,MAX_RSA_MODULUS_LEN, stream);
  fputs("    exp=",stream);
  DumpBigNum(pubKey->exponent,MAX_RSA_MODULUS_LEN, stream);
}

/*--- function DumpPrivKey ------------------------------------------
 *
 */
void
DumpPrivKey(privKey, stream)
R_RSA_PRIVATE_KEY *privKey;
FILE *stream;
{
  fprintf(stream,"Dump of %d bit private key:\n",privKey->bits);
  fputs(" mod   =",stream);
  DumpBigNum(privKey->modulus,MAX_RSA_MODULUS_LEN, stream);
  fputs(" pubExp=",stream);
  DumpBigNum(privKey->publicExponent,MAX_RSA_MODULUS_LEN, stream);
  fputs(" exp   =",stream);
  DumpBigNum(privKey->exponent,MAX_RSA_MODULUS_LEN, stream);
  fputs(" prime1=",stream);
  DumpBigNum(privKey->prime[0],MAX_RSA_PRIME_LEN, stream);
  fputs(" prime2=",stream);
  DumpBigNum(privKey->prime[1],MAX_RSA_PRIME_LEN, stream);
  fputs(" prExp1=",stream);
  DumpBigNum(privKey->primeExponent[0],MAX_RSA_PRIME_LEN, stream);
  fputs(" prExp2=",stream);
  DumpBigNum(privKey->primeExponent[1],MAX_RSA_PRIME_LEN, stream);
  fputs(" coeffi=",stream);
  DumpBigNum(privKey->coefficient,MAX_RSA_PRIME_LEN, stream);
}

/*--- function DumpBigNum -------------------------------------------
 *
 */
void
DumpBigNum(bigNum, numLen, stream)
unsigned char *bigNum;
int numLen;
FILE *stream;
{
  char buf[4];
  int j, bytesonline=0;
  
  for(j=0; j<numLen && !bigNum[j]; j++);
  for(; j<numLen; j++) {
    BinToHex(bigNum+j,1,buf);
    if(++bytesonline >= 32) {
      fputs("\n        ",stream);
      bytesonline=1;
    }
    fputs(buf,stream);
  }
  fputs("\n",stream);
}

/* Look in the public key source files for all records with the smartName and
     append the contents of each CertificateInfo field to the certs list.
   This will ASCII decode the CertificateInfo fields.
   RIPEMInfo is only used for debug and debugStream.
 */
char *GetCertsBySmartname (ripemDatabase, certs, smartName, ripemInfo)
RIPEMDatabase *ripemDatabase;
TypList *certs;
char *smartName;
RIPEMInfo *ripemInfo;
{
  BOOL gotCert;
  TypFile *typFile;
  char *errorMessage;
  unsigned char *newCert;
  unsigned int i, newCertLen;

  for (i = 0; i < MAX_KEY_SOURCES; ++i) {
    /* Only look in files for now. */
    if (ripemDatabase->pubKeySource.origin[i] == KEY_FROM_FILE) {
      FORLIST(&ripemDatabase->pubKeySource.filelist);
        typFile = (TypFile *)dptr;
        if (!typFile->stream)
          return ("Public key file is not open.");

        /* Rewind the file */
        fseek (typFile->stream, 0L, 0);

        /* Look for certificates until end of file.
         */
        do {
          if ((errorMessage = GetKeyBytesFromFile
               (smartName, typFile, CERT_INFO_FIELD, &gotCert, &newCert,
                &newCertLen, ripemInfo)) != (char *)NULL)
            return (errorMessage);

          if (!gotCert)
            /* Finished with this record.  Try the next. */
            continue;

          /* Append the newly allocated cert to the list. */
          if ((errorMessage = AddToList
               ((TypListEntry *)NULL, newCert, newCertLen, certs))
              != (char *)NULL) {
            /* error, so free the newCert */
            free (newCert);
            return (errorMessage);
          }
        } while (!feof (typFile->stream));
      ENDFORLIST;
    }
  }

  return ((char *)NULL);
}

char *WriteCert (certDER, certDERLen, pubOutStream)
unsigned char *certDER;
unsigned int certDERLen;
FILE *pubOutStream;
{
  unsigned int smartNameIndex;
  CertificateStruct *certStruct = (CertificateStruct *)NULL;
  char *errorMessage = (char *)NULL;

  /* For error, break to end of do while (0) block. */
  do {
    /* Allocate the certStruct on the heap because it's big. */
    if ((certStruct = (CertificateStruct *)malloc
         (sizeof (*certStruct))) == (CertificateStruct *)NULL) {
      errorMessage = ERR_MALLOC;
      break;
    }

    if (DERToCertificate
        (certDER, certStruct, (CertFieldPointers *)NULL) < 0) {
      errorMessage = "Can't decode certificate while saving it";
      break;
    }

    /* We are going to user the smart name in the "username" field.
     */
    GetDNSmartNameIndex (&smartNameIndex, &certStruct->subject);
    fprintf
      (pubOutStream, "%s %s\n", USER_FIELD,
       certStruct->subject.AVAValues[smartNameIndex]);

    fprintf (pubOutStream, "%s\n", CERT_INFO_FIELD);
    CodeAndWriteBytes (certDER, certDERLen, " ", pubOutStream);
  } while (0);

  free (certStruct);
  return (errorMessage);
}

